/* pty.c */

#include "pty.h"
#include "control.h"


TTYInfo     OriginalTTY;

#ifdef dBUG
FILE*       bugfile;
#endif

char*       SHELLNAME = "/bin/csh";

extern int  errno;

static CTerminal*   Root = NIL;


/* Close a PTY access channel and kill the process running on it. */
void          DeletePty(CTerminal* TheTerminal)
  {
    CTerminal*    Scan;
    CTerminal*    Lag;
    int           error;
    short         SessionNum;

    APRINT(("+DeletePty\n"));
    /* removing login entry */
    DeleteUtmpEntry(TheTerminal->UtmpIndex);
    /* closing file descriptor */
    if (TheTerminal->PID != 0)
      {
        int         Result;

        Result = kill(TheTerminal->PID,SIGKILL); /* unconditionally terminate the process */
      }
    error = close(TheTerminal->MasterDesc);

    while (TheTerminal->OldestBuffer != NIL)
      {
        PtyBufRec*    Temp;

        Temp = TheTerminal->OldestBuffer;
        TheTerminal->OldestBuffer = (TheTerminal->OldestBuffer)->NewerNext;
        free (Temp);
      }

    /* removing entry from linked list */
    Scan = Root;
    Lag = NIL;
    while ((Scan != TheTerminal) && (Scan != NIL))
      {
        Lag = Scan;
        Scan = Scan->Next;
      }
    ERROR(Scan==NIL,{APRINT(("fatal internal linked list error\n"));exit(1);})
    SessionNum = 0;
    while ((SessionNum < MaxNumSessions) && (Handlers[SessionNum] != TheTerminal))
      {
        SessionNum += 1;
      }
    ERROR(SessionNum>=MaxNumSessions,{APRINT(("fatal internal Handlers table error\n"));exit(1);})
    Handlers[SessionNum] = NIL;

    if (Lag == NIL)
      {
        Root = TheTerminal->Next;
      }
     else
      {
        Lag->Next = TheTerminal->Next;
      }
    free(TheTerminal);
    APRINT(("-DeletePty\n"));
  }


/* Attempts to create a new PTY.  Returns a pointer to it's memory cell if */
/* successful, or returns NIL if it failed. */
CTerminal*    NewPty(void)
  {
    char        Name[] = "/dev/pty..";
    char        Alpha[] = "pqrstuvwxyzPQRSTUVWXYZ"; /* pooh */
    char        Hex[16] = "0123456789abcdef";
    short       AlphaScan;
    short       NumberScan;
    int         MasterSideDesc;
    int         SlaveSideDesc;
    CTerminal*  Ref;
    int         PID;
    TTYInfo     NewTTY;
    int         error;

    APRINT(("+NewPty\n"));
    /* first, we try to find a PTY that's available. */
    for (AlphaScan = 0; AlphaScan < strlen(Alpha); AlphaScan += 1)
      {
        for (NumberScan = 0; NumberScan < strlen(Hex); NumberScan += 1)
          {
            Name[8] = Alpha[AlphaScan];
            Name[9] = Hex[NumberScan];
            APRINT(("Trying PTY:  '%s'\n",Name));
            MasterSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/ | O_NDELAY,0);
            if (MasterSideDesc >= 0) /* no error */
              {
                Name[5] = 't'; /* try to open the slave side, with blocking */
                /* should we ask for exclusive use here? */
                APRINT(("Trying TTY:  '%s'\n",Name));
                SlaveSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/,0);
                if (SlaveSideDesc >= 0) /* no error */
                  {
                    APRINT(("PTY/TTY successfully opened\n"));
                    goto WeHavePtyPoint;
                  }
                Name[5] = 'p';
              }
          }
      }
    APRINT(("-NewPty failed, no openable PTYs\n"));
    return NIL;  /* return NIL if we failed to open a PTY */

   WeHavePtyPoint:
    APRINT(("Master side descriptor = %d\n",MasterSideDesc));
    APRINT(("Slave side descriptor = %d\n",SlaveSideDesc));
    /* now we have to configure it */
    /* first, closing slave side */
    error = close(SlaveSideDesc);
    /* Now, setting the modes for the server's side (master) */
    NewTTY = OriginalTTY;
    APRINT(("Attempting to set Master Side to nonblocking raw mode\n"));
    MakeSpiffyTTY(&NewTTY);
    SetTTYInfo(MasterSideDesc,&NewTTY);
    APRINT(("Master Side should now be in nonblocking raw mode\n"));
    /* now, trying to make us exclusive owners of the TTY */
#if 0
    /* This stuff probably doesn't work unless we are installed as ROOT, and */
    /* nobody trusts this program that much, not even me! */
    error = seteuid(getuid(),geteuid());
    error = chown(Name,getuid(),getgid());
    ERROR(error==-1,APRINT(("chown on slave TTY failed %d\n",errno)));
    error = chmod(Name,0600);
    ERROR(error==-1,APRINT(("chmod on slave TTY failed %d\n",errno)));
#endif

    /* now we start the new process */
    PID = fork();
    if (PID == -1)
      {
        /* if we were unable to create the process, then we fail. */
        error = close(MasterSideDesc);
        APRINT(("-NewPty fork failed\n"));
        return NIL;
      }
    if (PID != 0)
      {
        /* master side */
#if 0
        error = setreuid(geteuid(),geteuid());
#endif
        Ref = (void*) malloc(sizeof(CTerminal));
        Ref->Next = Root;
        Root = Ref;
        Ref->MasterDesc = MasterSideDesc; /* remember file descriptor */
        Ref->PID = PID; /* remember process ID */
        Ref->OldestBuffer = NIL;
        Ref->NewestBuffer = NIL;
        Ref->UtmpIndex = MakeUtmpEntry(Name);
        APRINT(("-NewPty master side, success\n"));
        return Ref;
      }
     else
      {
        int                 error;
        int                 MyPid;
        struct itimerval    Timer;
        struct itimerval    TimerOld;
        char*               stupidparameters[2];
        char*               shell;
        int                 TempFD;

        /* slave side */
        APRINT(("Slave side:  fork entry complete\n"));
        close(MasterSideDesc); /* ya wanna make somethin' of it? */
        close(0);
        close(1);
        close(2);
        /* disabling timer */
        Timer.it_interval.tv_sec = 0;
        Timer.it_interval.tv_usec = 0;
        Timer.it_value.tv_sec = 0;
        Timer.it_value.tv_usec = 0;
        setitimer(ITIMER_REAL,&Timer,&TimerOld);
        /* restoring default signals */
        signal(SIGALRM,SIG_DFL);
        signal(SIGQUIT,SIG_DFL);

#if 0
        /* I have no idea why this is done, but everyone else does it. */
        setuid(getuid());
        setgid(getgid());
#endif

#if defined(ultrix) || defined(sparc)
        TempFD = open("/dev/tty",O_RDWR);
        error = ioctl(TempFD,TIOCNOTTY,(char*)0);
        ERROR(error < 0,APRINT(("slave side:  ioctl(TempFD,TIOCNOTTY,(char*)0) r
eturned %d, errno = %d\n",error,errno)));
        error = ioctl(TempFD,TIOCSPGRP,&MyPid);
        ERROR(error < 0,APRINT(("slave side:  ioctl(TempFD,TIOCNOTTY,&MyPid) ret
urned %d, errno = %d for the second time\n",error,errno)));
        close(TempFD);
#elif defined (hpux)
        setsid();
#endif

        /* I sure hope when we open the TTY, it becomes our controlling terminal */
        SlaveSideDesc = open(Name,O_RDWR /*| O_NOCTTY*/,0);
        APRINT(("slave side:  file descriptor = %d, should be 0\n",SlaveSideDesc));
#if defined(ultrix) || defined(sparc)
        error = ioctl(SlaveSideDesc,TIOCEXCL,(void*)NULL);
        ERROR(error < 0,APRINT(("slave side:  ioctl(SlaveSideDesc,TIOCEXCL,(void*)NULL) returned %d, error %d\n",error,errno)));
#elif defined (hpux)
        SetTTYInfo(SlaveSideDesc,&OriginalTTY);
#endif

#if defined(ultrix) || defined(sparc)
        MyPid = getpid();
        error = ioctl(0,TIOCSPGRP,&MyPid); /* what does this do? */
        ERROR(error < 0,APRINT(("slave side:  ioctl(0,TIOCSPGRP,&MyPid) returned %d, error %d\n",error,errno)));
        setpgrp(0,MyPid);
#elif defined (hpux)
        setpgrp();
#endif

        /* install the terminal in our standard streams */
        /* the open installed the terminal into 0 */
        error = dup(SlaveSideDesc); /* install into 1 */
        ERROR(error < 0,APRINT(("slave side:  dup(SlaveSideDesc) returned %d, errno = %d\n",error,errno)));
        error = dup(SlaveSideDesc); /* install into 2 */
        ERROR(error < 0,APRINT(("slave side:  dup(SlaveSideDesc) returned %d, errno = %d\n",error,errno)));

#if 0
        TempFD = open("/dev/tty",O_RDWR);
        error = ioctl(TempFD,TIOCNOTTY,(char*)0);
        ERROR(error < 0,APRINT(("slave side:  ioctl(TempFD,TIOCNOTTY,(char*)0) returned %d, errno = %d\n",error,errno)));
        error = ioctl(TempFD,TIOCSPGRP,&MyPid);
        ERROR(error < 0,APRINT(("slave side:  ioctl(TempFD,TIOCNOTTY,&MyPid) returned %d, errno = %d for the second time\n",error,errno)));
        close(TempFD);
#endif

        /* shelling out to the shell in the shell variable or the default shell */
        /* in SHELLNAME if there is no shell variable */
        shell = getenv("SHELL");
        if (shell == NIL)
          {
            APRINT(("slave side:  SHELL environment variable doesn't exist\n"));
            shell = SHELLNAME;
          }
        APRINT(("slave side:  execv shell '%s'\n",shell));
        stupidparameters[0] = shell; /* shell */
        stupidparameters[1] = NIL;
        APRINT(("execv'ing on slave side\n"));
        execv(SHELLNAME,stupidparameters);  /* bring up a shell */
        APRINT(("execv on slave side failed %d\n",errno));
        exit(1); /* abort, should never happen, but it might, nahh it couldn't... */
      }
  }


/* Send data to a PTY */
void          DispatchPty(CTerminal* TheTerminal, char* TheData, long NumBytes)
  {
    int           error;
    PtyBufRec*    Temp;

    while (NumBytes > 0)
      {
        /* making and linking buffers onto the output queue */
        Temp = (void*) malloc(sizeof(PtyBufRec));
        if (NumBytes > PtyBufSize)
          {
            Temp->NumBytesInBuffer = PtyBufSize;
          }
         else
          {
            Temp->NumBytesInBuffer = NumBytes;
          }
        memcpy(&(Temp->Data[0]),TheData,Temp->NumBytesInBuffer);
        /* establising link */
        Temp->NewerNext = NIL;
        if (TheTerminal->NewestBuffer == NIL)
          {
            TheTerminal->NewestBuffer = Temp;
            TheTerminal->OldestBuffer = Temp;
          }
         else
          {
            TheTerminal->NewestBuffer->NewerNext = Temp;
            TheTerminal->NewestBuffer = Temp;
          }
        TheData += Temp->NumBytesInBuffer;
        NumBytes -= Temp->NumBytesInBuffer;
      }
  }


/* Read data from PTYs.  Also detect terminated processes. */
void          PtyIdle(void)
  {
    CTerminal*    Scan;
    char          Buffer[256];
    int           NumBytes;
    short         Item;
    pid_t         PID;
#if defined(ultrix) || defined(sparc)
    union wait    Status;
#elif defined (hpux)
    int           Status;
#endif

    Scan = Root;
    while (Scan != NIL)
      {
        Item = 0;
        /* identifying which session it is */
        while (Handlers[Item] != Scan)
          {
            Item += 1;
          }
        /* seeing if there is any data to read */
        if (((Buffers[Item].DataEnd - Buffers[Item].DataStart) & BuffMask)
          + 255 < BuffSize - 1)
          {
            /* there is room to write data to the local buffer. */
            NumBytes = read(Scan->MasterDesc,Buffer,255);
            if (NumBytes > 0)
              {
                /* some data was read without error, we can send it. */
                ReceiveDataFromLocalTerminal(Buffer,NumBytes,Scan);
              }
          } /* end of if there's room in our local buffer */
        /* seeing if there is any data to write */
        if (Scan->OldestBuffer != NIL)
          {
            NumBytes = write(Scan->MasterDesc,Scan->OldestBuffer->Data,
              Scan->OldestBuffer->NumBytesInBuffer);
            if (NumBytes > 0)
              {
                /* if data was successfully written, see how much */
                if (NumBytes == Scan->OldestBuffer->NumBytesInBuffer)
                  {
                    PtyBufRec*    Temp;

                    /* entire write was successful, we can dequeue */
                    Temp = Scan->OldestBuffer;
                    Scan->OldestBuffer = Scan->OldestBuffer->NewerNext;
                    if (Scan->OldestBuffer == NIL)
                      {
                        /* this happens when there was only one buffer */
                        Scan->NewestBuffer = NIL;
                      }
                    free(Temp);
                  }
                 else
                  {
                    int     Inxs;

                    /* the entire write was not successful, but some was */
                    for (Inxs = 0; Inxs < Scan->OldestBuffer->NumBytesInBuffer - NumBytes;
                      Inxs += 1)
                      {
                        Scan->OldestBuffer->Data[Inxs]
                          = Scan->OldestBuffer->Data[NumBytes + Inxs];
                      }
                    Scan->OldestBuffer->NumBytesInBuffer -= NumBytes;
                  }
              }
          } /* end of writing to the PTY */
        Scan = (*Scan).Next;
      } /* end of scan through all pty loop */
    PID = wait3(&Status,WNOHANG,NIL/*no usage*/);
    if (PID > 0)
      {
        CTerminal*    Check;

        Check = Root;
        while (Check != NIL)
          {
            if (Check->PID == PID)
              {
                if ((WIFEXITED(Status)) || WIFSIGNALED(Status))
                  {
                    /* the process exited, so don't kill it */
                    /* We don't do this if it was only WIFSTOPPED */
                    /* since we have to kill stopped processes ourself */
                    Check->PID = 0;
                  }
                TerminalWantsToDieCloseSession(Check);
                Check = NIL;
              }
             else
              {
                Check = (*Check).Next;
              }
          }
        /* if the process's record doesn't exist, then */
        /* we just fall through the loop.  this happens if it was killed */
        /* but not waited for elsewhere. */
      } /* end of if wait3's pid != 0 */
  }


void          WindowSizeChange(CTerminal* TheTerminal, short NewX, short NewY)
  {
    struct winsize    MyWindowSize;

    APRINT(("+WindowSizeChange\n"));
#ifdef TIOCSWINSZ
    MyWindowSize.ws_col = NewX;
    MyWindowSize.ws_row = NewY;
    MyWindowSize.ws_xpixel = 0;
    MyWindowSize.ws_ypixel = 0;
    ioctl(TheTerminal->MasterDesc,TIOCSWINSZ,&MyWindowSize);
#endif
    APRINT(("-WindowSizeChange\n"));
  }


/****************************************************************/
/* Getting and Setting and Getting and Setting and Getting TTY Stuff */


void    GetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    APRINT(("+GetTTYInfo\n"));
#if defined(ultrix) || defined(sparc)
    ioctl(TheFile,TIOCGETC,&(TheInfo->TCHARS));
    ioctl(TheFile,TIOCGETD,&(TheInfo->LDISC));
    ioctl(TheFile,TIOCGETP,&(TheInfo->TTYB));
    ioctl(TheFile,TIOCGLTC,&(TheInfo->LTCHARS));
    ioctl(TheFile,TIOCLGET,&(TheInfo->LMODE));
#elif defined (hpux)
    tcgetattr(TheFile,&(TheInfo->TERMIOS));
#endif
    TheInfo->fcntlFlags = fcntl(TheFile,F_GETFL,0);
    APRINT(("-GetTTYInfo\n"));
  }


void    SetTTYInfo(int TheFile, TTYInfo* TheInfo)
  {
    int     error;

    APRINT(("+SetTTYInfo\n"));
#if defined(ultrix) || defined(sparc)
    error = ioctl(TheFile,TIOCSETC,&(TheInfo->TCHARS));
    ERROR(error < 0,APRINT(("ioctl:  TIOCSETC returned %d, errno %d\n",error,errno)));
    error = ioctl(TheFile,TIOCSETD,&(TheInfo->LDISC));
    ERROR(error < 0,APRINT(("ioctl:  TIOCSETD returned %d, errno %d\n",error,errno)));
    error = ioctl(TheFile,TIOCSETP,&(TheInfo->TTYB));
    ERROR(error < 0,APRINT(("ioctl:  TIOCSETP returned %d, errno %d\n",error,errno)));
    error = ioctl(TheFile,TIOCSLTC,&(TheInfo->LTCHARS));
    ERROR(error < 0,APRINT(("ioctl:  TIOCSLTC returned %d, errno %d\n",error,errno)));
    error = ioctl(TheFile,TIOCLSET,&(TheInfo->LMODE));
    ERROR(error < 0,APRINT(("ioctl:  TIOCLSET returned %d, errno %d\n",error,errno)));
#elif defined (hpux)
    error = tcsetattr(TheFile,TCSANOW,&(TheInfo->TERMIOS));
    ERROR(error < 0,APRINT(("tcsetattr:  returned %d, errno %d\n",error,errno)));
#endif
    error = fcntl(TheFile,F_SETFL,TheInfo->fcntlFlags);
    ERROR(error < 0,APRINT(("fcntl:  F_SETFL returned %d, errno %d\n",error,errno)));
#if 0
    error = ioctl(TheFile,TIOCEXCL,0);
    ERROR(error==-1,APRINT(("SetTTYInfo ioctl for exclusive use error %d\n",errno)));
#endif
    APRINT(("-SetTTYInfo\n"));
  }


void    MakeSpiffyTTY(TTYInfo* TheInfo)
  {
    APRINT(("+MakeSpiffyTTY\n"));
#if defined(ultrix) || defined(sparc)
    TheInfo->TTYB.sg_flags = RAW | ANYP;
    TheInfo->TCHARS.t_intrc = -1;
    TheInfo->TCHARS.t_quitc = -1;
    TheInfo->TCHARS.t_startc = -1;
    TheInfo->TCHARS.t_stopc = -1;
    TheInfo->LTCHARS.t_suspc = -1;
    TheInfo->LTCHARS.t_dsuspc = -1;
    TheInfo->LTCHARS.t_flushc = -1;
    TheInfo->LTCHARS.t_lnextc = -1;
#elif defined (hpux)
    TheInfo->TERMIOS.c_iflag = IGNBRK | IGNPAR;
    TheInfo->TERMIOS.c_oflag = 0;
    TheInfo->TERMIOS.c_cflag = CLOCAL | CS8 | B38400;
    TheInfo->TERMIOS.c_lflag = 0;
    TheInfo->TERMIOS.c_cc[VTIME] = 0;
    TheInfo->TERMIOS.c_cc[VMIN] = 0;
#endif
    TheInfo->fcntlFlags |= O_NDELAY;
    APRINT(("-MakeSpiffyTTY\n"));
  }


/****************************************************************/
/* Fun with UTMP! */


static int      UtmpFile;
static short    UtmpValid = False;


void    OpenUtmpFile(void)
  {
    APRINT(("+OpenUtmpFile\n"));
#ifdef USE_UTMP
    UtmpFile = open("/etc/utmp",O_RDWR,0);
    if (UtmpFile != -1)
      {
        UtmpValid = True; /* opened successfully, so we can use it */
      }
#endif
    APRINT(("-OpenUtmpFile\n"));
  }


/* return the index number of the TTYName in the TTY list. */
/* 0 = not found, 1 = first one, 2 = second one... */
int     FindUtmpEntry(char* TTYName)
  {
    struct ttyent*    Thingie;
    char*             SpiffName;
    int               Index;

    APRINT(("+FindUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        SpiffName = (char*)rindex(TTYName,'/');
        if (SpiffName != NIL)
          {
            SpiffName += 1; /* step past '/' to first character of name */
          }
         else
          {
            SpiffName = TTYName;
          }
        setttyent(); /* reset TTY scanning thing */
        Index = 0;
       Loop:
        Index += 1;
        Thingie = getttyent(); /* get a tty thing */
        if (Thingie == NIL)
          {
            APRINT(("-FindUtmpEntry found nothing\n"));
            return 0; /* means we found nothing */
          }
        if (strcmp(Thingie->ty_name,SpiffName) == 0)
          {
            APRINT(("-FindUtmpEntry %d\n",(int)Index));
            /* names are equal, so we've found our man */
            return Index;
          }
        goto Loop;
      }
#endif
    APRINT(("-FindUtmpEntry found nothing\n"));
    return 0; /* this gets returned if we couldn't access file */
  }


int     GetUtmpEntry(int Index, struct utmp* TheUtmpEntry)
  {
    int     BytesRead = 0;

    APRINT(("+GetUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        /* first, go to the location in UTMP */
        lseek(UtmpFile,(long)(Index * sizeof(struct utmp)),0);
        BytesRead = read(UtmpFile,(void*)TheUtmpEntry,sizeof(struct utmp));
      }
#endif
    APRINT(("-GetUtmpEntry %d\n",(int)BytesRead));
    return (BytesRead > 0); /* true = worked, false = failed */
  }


void    PutUtmpEntry(int Index, struct utmp* TheUtmpEntry)
  {
    int     BytesWritten;

    APRINT(("+PutUtmpEntry\n"));
#ifdef USE_UTMP
    if (UtmpValid)
      {
        lseek(UtmpFile,(long)(Index * sizeof(struct utmp)),0);
        BytesWritten = write(UtmpFile,(void*)TheUtmpEntry,sizeof(struct utmp));
      }
#endif
    APRINT(("-PutUtmpEntry\n"));
  }


void    DeleteUtmpEntry(int Index)
  {
    struct utmp   ZapEntry;
    int           Scan;

    APRINT(("+DeleteUtmpEntry\n"));
#ifdef USE_UTMP
    for (Scan = 0; Scan < sizeof(struct utmp); Scan += 1)
      {
        /* portable? nahhh.. */
        ((char*)(&ZapEntry))[Scan] = 0;
      }
    PutUtmpEntry(Index,&ZapEntry);
#endif
    APRINT(("-DeleteUtmpEntry\n"));
  }


extern struct utmp* RootUtmpEntry; /* from main.c */

int     MakeUtmpEntry(char* Name)
  {
    struct utmp   OurUtmpEntry;
    int           Where;

    APRINT(("+MakeUtmpEntry\n"));
#ifdef USE_UTMP
    Where = FindUtmpEntry(Name);
    if (UtmpValid && (Where != 0))
      {
        if (rindex(Name,'/') == 0)
          {
            strncpy(OurUtmpEntry.ut_line,Name,8);
          }
         else
          {
            strncpy(OurUtmpEntry.ut_line,(rindex(Name,'/') + (char*)1),8);
          }
        strncpy(OurUtmpEntry.ut_name,RootUtmpEntry->ut_name,8);
        strncpy(OurUtmpEntry.ut_host,"Unknown",8);
        time(&OurUtmpEntry.ut_time);
        PutUtmpEntry(Where,&OurUtmpEntry);
        APRINT(("-MakeUtmpEntry %d\n",(int)Where));
        return Where;
      }
#endif
    APRINT(("-MakeUtmpEntry 0\n"));
    return 0;
  }


/****************************************************************/


#ifdef dBUG
void    AuditPrint(char* First, ...)
  {
    va_list     ap;

    va_start(ap,First);
    vfprintf(bugfile,First,ap);
    va_end(ap);
    fflush(bugfile);
  }
#endif
